package main

import (
	"context"
	"log"
	"time"

	"git.netflux.io/rob/clipper/config"
	"git.netflux.io/rob/clipper/generated/store"
	"git.netflux.io/rob/clipper/media"
	"git.netflux.io/rob/clipper/server"
	awsconfig "github.com/aws/aws-sdk-go-v2/config"
	"github.com/aws/aws-sdk-go-v2/credentials"
	"github.com/aws/aws-sdk-go-v2/service/s3"
	"github.com/jackc/pgx/v4/pgxpool"
	"github.com/kkdai/youtube/v2"
)

const (
	DefaultHTTPBindAddr = "0.0.0.0:8888"
	DefaultTimeout      = 600 * time.Second
)

func main() {
	ctx := context.Background()

	config, err := config.NewFromEnv()
	if err != nil {
		log.Fatalf("error loading config: %v", err)
	}

	// Create a store
	dbConn, err := pgxpool.Connect(ctx, config.DatabaseURL)
	if err != nil {
		log.Fatal(err)
	}
	store := store.New(dbConn)

	// Create an Amazon S3 service s3Client
	cfg, err := awsconfig.LoadDefaultConfig(
		ctx,
		awsconfig.WithCredentialsProvider(
			credentials.NewStaticCredentialsProvider(config.AWSAccessKeyID, config.AWSSecretAccessKey, ""),
		),
		awsconfig.WithRegion(config.AWSRegion),
	)
	if err != nil {
		log.Fatal(err)
	}
	s3Client := s3.NewFromConfig(cfg)

	// Create an Amazon S3 presign client
	s3PresignClient := s3.NewPresignClient(s3Client)

	// Create a Youtube client
	var youtubeClient youtube.Client

	serverOptions := server.Options{
		Config:        config,
		BindAddr:      DefaultHTTPBindAddr,
		Timeout:       DefaultTimeout,
		Store:         store,
		YoutubeClient: &youtubeClient,
		S3API: media.S3API{
			S3Client:        s3Client,
			S3PresignClient: s3PresignClient,
		},
	}

	log.Fatal(server.Start(serverOptions))
}
